@@ -1,5 +1,5 @@
 use std::collections::{BTreeMap, HashMap};
-use std::convert::{TryFrom, TryInto};
+use std::convert::TryFrom;
 use std::fmt;
 use std::iter::FromIterator;
 use std::sync::Arc;
@@ -11,7 +11,7 @@ use log::debug;
 
 use tc_btree::{BTreeFile, BTreeInstance, BTreeType, Node};
 use tc_error::*;
-use tc_transact::fs::{Dir, File, Persist, Restore};
+use tc_transact::fs::{CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{Transact, Transaction, TxnId};
 use tc_value::Value;
 use tcgeneric::{label, Id, Instance, Label, TCTryStream, Tuple};
@@ -278,8 +278,6 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> ReadOnly<F, D, Txn> {
             .create_file_tmp(*txn.id(), BTreeType::default())
             .await?;
 
-        let file = file.try_into()?;
-
         let source_schema: IndexSchema = (source.key().to_vec(), source.values().to_vec()).into();
 
         let (schema, btree) = if let Some(columns) = key_columns {
@@ -413,7 +411,6 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
             .create_file(txn_id, PRIMARY_INDEX.into(), BTreeType::default())
             .await?;
 
-        let primary_file = primary_file.try_into()?;
         let primary = Index::create(primary_file, schema.primary().clone(), txn_id).await?;
 
         let primary_schema = schema.primary();
@@ -434,8 +431,6 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
                         .create_file(txn_id, name.clone(), BTreeType::default())
                         .await?;
 
-                    let file = file.try_into()?;
-
                     Self::create_index(file, primary_schema, column_names, txn_id)
                         .map_ok(move |index| (name, index))
                         .await
@@ -944,6 +939,7 @@ impl<F: File<Node> + Transact, D: Dir, Txn: Transaction<D>> Transact for TableIn
 impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Persist<D, Txn> for TableIndex<F, D, Txn>
 where
     F: TryFrom<D::File, Error = TCError>,
+    <D as Dir>::FileClass: From<BTreeType> + Send,
 {
     type Schema = TableSchema;
     type Store = D;
@@ -958,7 +954,7 @@ where
             .await?
             .ok_or_else(|| TCError::internal("cannot load Table: primary index is missing"))?;
 
-        let primary = Index::load(txn, schema.primary().clone(), file.try_into()?).await?;
+        let primary = Index::load(txn, schema.primary().clone(), file).await?;
 
         let mut auxiliary = Vec::with_capacity(schema.indices().len());
         for (name, columns) in schema.indices() {
@@ -968,7 +964,7 @@ where
 
             let index_schema = schema.primary().auxiliary(columns)?;
 
-            let index = Index::load(txn, index_schema, file.try_into()?).await?;
+            let index = Index::load(txn, index_schema, file).await?;
             auxiliary.push((name.clone(), index));
         }
 
@@ -1012,6 +1008,33 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Restore<D, Txn> for TableIndex<
     }
 }
 
+#[async_trait]
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>, I: TableInstance<F, D, Txn>> CopyFrom<D, Txn, I>
+    for TableIndex<F, D, Txn>
+where
+    F: TryFrom<D::File, Error = TCError>,
+    <D as Dir>::FileClass: From<BTreeType> + Send,
+{
+    async fn copy_from(source: I, dir: D, txn_id: TxnId) -> TCResult<Self>
+    where
+        I: 'async_trait,
+    {
+        let schema = source.schema();
+        let key_len = schema.primary().key().len();
+        let table = Self::create(schema, &dir, txn_id).await?;
+
+        let rows = source.rows(txn_id).await?;
+
+        rows.map_ok(|mut row| (row.drain(..key_len).collect(), row))
+            .map_ok(|(key, values)| table.upsert(txn_id, key, values))
+            .try_buffer_unordered(num_cpus::get())
+            .try_fold((), |(), ()| future::ready(Ok(())))
+            .await?;
+
+        Ok(table)
+    }
+}
+
 impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<TableIndex<F, D, Txn>> for Table<F, D, Txn> {
     fn from(table: TableIndex<F, D, Txn>) -> Self {
         Self::Table(table)
